﻿using System.IO;
using System;
using System.Collections.Generic;
using System.ServiceProcess;
using System.Threading.Tasks;
using System.Threading;
using System.Timers;
using System.ServiceModel;
using OccupationMonitorLib.Drawers;
using OccupationMonitorLib.Processors;
using OccupationMonitorLib.DataAccess;
using OccupationMonitorLib.MiniZappa;
using OccupationMonitorLib.TextOutput;
using OccupationMonitorLib;

namespace OccupationMonitor
{
    public partial class OccupationMonitorService : ServiceBase
    {
        static IOccupationMonitor _occupationMonitor;

        static ErrorLogger _errLog;
        static MiniZappaNewClient _zappaClient;
        static Thread _currentThread = null;
        static ServiceHost _host;
        System.Timers.Timer _timer;

        public OccupationMonitorService()
        {
            InitializeComponent();
        }

        public void Start()
        {
            try
            {
                _errLog = new ErrorLogger(Config.LoggerErrorLogPath, DateTime.Now.ToString("yyyy-MM-dd"), Config.EventLogSource);
                _occupationMonitor = new OccupationMonitorLib.OccupationMonitor();
                _zappaClient = new MiniZappaNewClient();
                
                _timer = new System.Timers.Timer(Config.TimerIntervalsMiliSecs);
                _timer.Elapsed += new ElapsedEventHandler(OnElapsed);
                _timer.Start(); 

                Uri addr = new Uri(Config.ServiceHostAddress);
                Type serviceType = typeof(OccupationMonitorLib.OccupationMonitor);
                _host = new ServiceHost(serviceType, addr);

                _host.Open();
            }
            catch (Exception ex)
            {
                _errLog.LogException(ex);
                this.ExitCode = 13816; //error code: "An unknown error has occurred."
                this.Stop();
            }
        }

        protected override void OnStart(string[] args)
        {
            Start();
        }

        protected override void OnStop()
        {
            CloseZappaClient();

            try
            {
                if (_host != null) _host.Close(); //su situacie kedy host moze byt null
            }
            catch (Exception)
            {
                if (_host != null) _host.Abort();
            }
        }

        private void OnElapsed(object source, ElapsedEventArgs e)
        {
            try
            {
                //kontroluje, ci vlakno ktore bezalo skor, uz skoncilo - ked neskoncilo, tak ho necha dokoncit a nove sa nespusti
                if (_currentThread == null
                    || _currentThread.ThreadState.HasFlag(System.Threading.ThreadState.Aborted)
                    || _currentThread.ThreadState.HasFlag(System.Threading.ThreadState.Stopped))
                {
                    _currentThread = Thread.CurrentThread;
                    _currentThread.IsBackground = true;
                    Thread thrd = new Thread(ConnectToZappaAndProcess);
                    thrd.IsBackground = true;
                    thrd.Start();
                    if (!thrd.Join(TimeSpan.FromSeconds(Config.LocalitiesProcessingTimeoutSecs)))//ked sa vlakno,ktore loguje, zasekne na dlhisu dobu, ukonci sa - za chvilu timer vyvola udalost, kedy to skusi znova
                    {
                        thrd.Abort();
                        Console.WriteLine("Timedout!");
                        _currentThread = null;
                        throw new System.TimeoutException("Timeout for processing localities expired!");
                    }
                    _currentThread = null;
                }
                else
                    Console.WriteLine("Previous job is not finished yet");
            }
            catch (Exception ex)
            {
                _errLog.LogException(ex);
            }
        }

        private void ConnectToZappaAndProcess()
        {
            try
            {
                Console.WriteLine("Connecting to minizappa");
                long[] aml = _zappaClient.GetAvailableMapLocalities();//vrati idcka dostupnych ucebni
                Console.WriteLine("MapLocalities available at: {0}", DateTime.Now);
                ProcessLocalities(aml);
            }
            catch (ThreadAbortException thrEx) //ked sa abortne vlakno, v ktorom bezi tato metoda, tak to v tomto vlakne vyhodi vynimku
            {
                _errLog.LogException(thrEx);
                Thread.ResetAbort();
            }
            catch (System.TimeoutException ex)
            {
                _errLog.LogException(ex);
                Recover();
            }
            catch (CommunicationException ex) 
            {
                _errLog.LogException(ex);
                Recover();
            }
            catch (Exception ex)
            {
                _errLog.LogException(ex);
            }
        }

        private void ProcessLocalities(long[] availableLocalities)
        {
            MapLocality ml;//n-arny strom, uzly su studovne, listy su pocitace
            Dictionary<long, WorkstationStatus> ws;//poskytuje stav pocitacov (id_pc, aktualny_stav)

            for (int i = 0; i < availableLocalities.Length; i++)
            {
                try 
                {
                    ml = _zappaClient.GetMaplocality(availableLocalities[i]);
                    ws = _zappaClient.GetMapLocalityWorkstationStatuses(availableLocalities[i]);

                    Console.WriteLine("Processing locality: {0} locality id: {1}", ml.Name, ml.Id);
                    if (ml != null && ws != null)
                    {
                        _occupationMonitor.LogState(ml, ws);//ulozi aktualny stav pocitacov
                        _occupationMonitor.GenerateImage(ml, ws);

                        try
                        {
                         _occupationMonitor.GenerateCsv(ml, ws);
                        }
                        catch (IOException ex) //zapisuje sa tam do suboru, subor moze citat IIS
                        {
                            _errLog.LogException(ex);
                        }
                    }
                }
                catch (System.TimeoutException ex)
                {
                    _errLog.LogException(ex);
                    Recover();
                }
                catch (CommunicationException ex)
                {
                    _errLog.LogException(ex);
                    Recover();
                }            
            }
            ml = null;
            ws = null;
        }

        private void Recover()
        {
            CloseZappaClient();
            GC.Collect(GC.MaxGeneration, GCCollectionMode.Forced);
            _zappaClient = new MiniZappaNewClient();
        }

        private void CloseZappaClient()
        {
            if (_zappaClient != null)  //su situacie kedy zappa client moze byt null
            {
                try
                {
                    _zappaClient.Close();
                }
                catch (Exception ex)
                {
                    _errLog.LogException(ex);
                }
                _zappaClient.Abort();
            }
            _zappaClient = null;
        }
    }
}
