﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Drawing;
using System.IO;
using System.ServiceModel;
using System.Configuration;
using OccupationMonitorLib.DataObjects;
using OccupationMonitorLib.Drawers;
using OccupationMonitorLib.DrawingStrategies;
using OccupationMonitorLib.ProcessingStrategies;
using OccupationMonitorLib.Processors;
using OccupationMonitorLib.MiniZappa;
using OccupationMonitorLib.Enums;
using OccupationMonitorLib.DataAccess;
using OccupationMonitorLib.TextOutput;


namespace OccupationMonitorLib
{
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.PerCall)]
    public class OccupationMonitor : IOccupationMonitor
    {
        ErrorLogger logger = new ErrorLogger(Config.OccServiceErrorLogPath, "Querry Errors " + DateTime.Now.ToString("yyyy-MM-dd"), Config.EventLogSource);

        public OccupationMonitor() 
        {
            CreateDirectories(Config.OccImagesPath, Config.LoggerErrorLogPath, Config.OccServiceErrorLogPath, Config.OccCsvPath);
            ImgsPaths.Initialize(Config.OccImagesPath);
            ConnectionSingleton.Instance.ConnectionString = Config.OracleConnString;
        }

        #region PublicMethods
        public byte[] GetImagesFromFile(long roomId)
        {
            try
            {
                string path = Config.OccImagesPath + roomId + ".bmp";
                lock (ImgsPaths.GetPath(path))
                {
                    using (Image img = Image.FromFile(path))
                    {
                        return ConvertToByte(img);
                    }
                }
            }
            catch (FileNotFoundException ex)
            {
                logger.LogException(ex);
                throw new FaultException("File not found probably due to invalid input parameter. Check, whether the room with given id exists: " + roomId);
            }
            catch (ArgumentException ex)
            {
                logger.LogException(ex);
                throw new FaultException("File not found probably due to invalid input parameter. Check, whether the room with given id exists: " + roomId);
            }
            catch (Exception ex)
            {
                logger.LogException(ex);
                throw new FaultException("Service unable to get file.");
            }
        }

        public void GenerateImage(MapLocality ml, Dictionary<long, WorkstationStatus> ws)
        {
            OccupationDrawer drawerOcc = new OccupationDrawer(Config.PcBorderLength, Config.PcLength, Config.OccPenSize, Config.OccLegendHeight);

            lock (ImgsPaths.GetPath(String.Format("{0}{1}.bmp", Config.OccImagesPath, ml.Id)))
            {
                using (System.Drawing.Image bmp = drawerOcc.DrawOccupation(ml, ws)) //vykresli stav ucebne
                {
                    bmp.Save(String.Format("{0}{1}.bmp", Config.OccImagesPath, ml.Id));
                    Console.WriteLine("Locality saved: {0}", ml.Id);
                }
            }
        }

        public void GenerateCsv(MapLocality ml, Dictionary<long, WorkstationStatus> ws)
        {
            CsvGenerator csvGenerator = new CsvGenerator();
            string csvData = csvGenerator.GetCsvData(ml, ws);

            File.WriteAllLines(String.Format("{0}{1}.csv", Config.OccCsvPath, ml.Id), new string[] { csvData });
        }

        public byte[] GetStatsImg(long roomId, StatFormat format)
        {
            try
            {
                IntervalsPreProcessors pre = new IntervalsPreProcessors();
                DateTime lastSunday = GetLastEndOfWeek();

                List<LogItem> logItems;
                IntervalsProcessor proc;
                StatsDrawer drawerStat;
                DateTime start;
                GetInfoForStats(roomId, format, lastSunday, out logItems, out proc, out drawerStat, out start);

                List<OccupationInterval> intervals = pre.ProcessLogs(logItems, lastSunday);
                float[] proccesed = proc.ProcessToPercents(GetPcsCount(roomId), intervals);

                using (Image img = drawerStat.DrawStatistics(proccesed, Config.StatsImgWidth, Config.StatsImgHeight, start, lastSunday))
                {
                    return ConvertToByte(img);
                }
            }
            catch (Exception ex)
            {
                logger.LogException(ex);
                if (ex.GetType() == typeof(FaultException)) throw ex;
                else throw new FaultException("Error occured while processing statistics.");
            }
        }

        public int GetFreePcsCount(long roomId)
        {
            MapLocality ml;
            Dictionary<long, WorkstationStatus> ws;
            try
            {
                using (MiniZappaNewClient sc = new MiniZappaNewClient())
                {
                    ml = sc.GetMaplocality(roomId);
                    ws = sc.GetMapLocalityWorkstationStatuses(roomId);
                }
            }
            catch (Exception ex)
            {
                logger.LogException(ex);
                throw new FaultException("Error occured while processing the request.");
            }

            if (ml != null)
            {
                int free = 0;
                for (int i = 0; i < ml.Studyrooms.Length; i++)
                {
                    MapObject room = ml.Studyrooms[i].RootMapObject;
                    free = CountFreeComputersInMapObject(ws, room);
                }
                return free;
            }
            throw new FaultException("Studyroom not found probably due to invalid input parameter. Check, whether the room with given id exists: " + roomId);
        }

        public int GetPcsCount(long roomId)
        {
            MapLocality ml;
            try
            {
                using (MiniZappaNewClient sc = new MiniZappaNewClient())
                {
                    ml = sc.GetMaplocality(roomId);
                }
            }
            catch (Exception ex)
            {
                logger.LogException(ex);
                throw new FaultException("Error occured while processing the request.");
            }

            if (ml != null)
            {
                int count = 0;
                foreach (MapStudyroom room in ml.Studyrooms)
                    count += GetChildrenCount(room.RootMapObject);
                return count;
            }
            throw new FaultException("Studyroom not found probably due to invalid input parameter. Check, whether the room with given id exists: " + roomId);
        }

        public Dictionary<long, string> GetAvailableLocalities()
        {
            try
            {
                Dictionary<long, string> locs = new Dictionary<long, string>();
                using (MiniZappaNewClient zappa = new MiniZappaNewClient())
                {
                    long[] aml = zappa.GetAvailableMapLocalities();
                    foreach (long id in aml)
                    {
                        try
                        {
                            locs.Add(id, zappa.GetMaplocality(id).Name);
                        }
                        catch (ArgumentException ex)
                        {
                            logger.LogException(ex);
                        }
                    }
                    return locs;
                }
            }
            catch (Exception ex)
            {
                logger.LogException(ex);
                throw new FaultException("Error occured while processing the request.");
            }
        }


        public void LogState(MapLocality ml, Dictionary<long, WorkstationStatus> ws) 
        {
            MapLocalityProcessor processor = new MapLocalityProcessor();
            processor.LogState(ml, ws);
        }




        #endregion

        #region PrivateMethods
        private int CountFreeComputersInMapObject(Dictionary<long, WorkstationStatus> ws, MapObject room)
        {
            int free = 0;
            for (int i = 0; i < room.Children.Length; i++)
            {
                if (room.Children[i].ObjectType == MapObjectType.Workstation)
                {
                    if (ws.ContainsKey((long)room.Children[i].Data))
                        if (ws[(long)room.Children[i].Data].StatusValue == WorkstationStatusEnum.Free) { free++; }
                }
                else if (room.Children[i].ObjectType == MapObjectType.Studyroom)
                    free += CountFreeComputersInMapObject(ws, room.Children[i]);
            }
            return free;
        }

        private int GetChildrenCount(MapObject mo)
        {
            int count = 0;
            foreach (MapObject child in mo.Children)
                if (child.ObjectType == MapObjectType.Workstation) count++;
                else if (child.ObjectType == MapObjectType.Studyroom) count += GetChildrenCount(child);
            return count;
        }

        private void GetInfoForStats(long roomId, StatFormat format, DateTime lastSunday, out List<LogItem> logItems, out IntervalsProcessor proc, out StatsDrawer drawerStat, out DateTime start)
        {
            MapLocalityProcessor logger = new MapLocalityProcessor();
            if (format == StatFormat.Hours)
            {
                start = GetMonthAgo(lastSunday);
                logItems = logger.GetLogs(roomId, GetMonthAgo(lastSunday), lastSunday);
                proc = new IntervalsProcessor(new ProcessToHours());
                drawerStat = new StatsDrawer(new DrawXAxisHours(), Config.StatsImgPadding, Config.StatsAxisPartLength, Config.StatsPartSizeCoefficient, Config.StatsPenSize);
            }
            else if (format == StatFormat.WeekDays)
            {
                start = GetMonthAgo(lastSunday);
                logItems = logger.GetLogs(roomId, GetMonthAgo(lastSunday), lastSunday);
                proc = new IntervalsProcessor(new ProcessToWeek());
                drawerStat = new StatsDrawer(new DrawXAxisWeek(), Config.StatsImgPadding, Config.StatsAxisPartLength, Config.StatsPartSizeCoefficient, Config.StatsPenSize);
            }
            else if (format == StatFormat.Months)
            {
                start = new DateTime(lastSunday.Year - 1, lastSunday.Month, lastSunday.Day, lastSunday.Hour, lastSunday.Minute, lastSunday.Second);
                logItems = logger.GetLogs(roomId, new DateTime(lastSunday.Year - 1, lastSunday.Month, lastSunday.Day, lastSunday.Hour, lastSunday.Minute, lastSunday.Second), lastSunday);
                proc = new IntervalsProcessor(new ProcessToMonths());
                drawerStat = new StatsDrawer(new DrawXAxisMonths(), Config.StatsImgPadding, Config.StatsAxisPartLength, Config.StatsPartSizeCoefficient, Config.StatsPenSize);
            }
            else throw new ArgumentException("invalid argument value: " + format);
        }

        //vrati datum poslednej nedele
        private static DateTime GetLastEndOfWeek()
        {
            DateTime now = DateTime.Now;
            int i = 1;
            while (now.AddDays(-i).DayOfWeek != DayOfWeek.Sunday) i++;
            DateTime toReturn = (now.AddDays(-i));
            return new DateTime(toReturn.Year, toReturn.Month, toReturn.Day, 23, 59, 59);

        }

        //vrati datum presne o mesiac skor
        private DateTime GetMonthAgo(DateTime date)
        {
            if (date.Month != 1)
            {
                if (date.Day > DateTime.DaysInMonth(date.Year, date.Month - 1))
                    return new DateTime(date.Year, date.Month - 1, DateTime.DaysInMonth(date.Year, date.Month - 1), date.Hour, date.Minute, date.Second);
                else
                    return new DateTime(date.Year, date.Month - 1, date.Day, date.Hour, date.Minute, date.Second);
            }
            else return new DateTime(date.Year - 1, 12, date.Day, date.Hour, date.Minute, date.Second);
        }

        //transformuje obrazok na pole bytov
        private byte[] ConvertToByte(Image img)
        {
            byte[] bytes;
            using (MemoryStream ms = new MemoryStream())
            {
                img.Save(ms, System.Drawing.Imaging.ImageFormat.Bmp);
                bytes = ms.ToArray();
            }
            return bytes;

        }

        private void GetDataFromZappa(long roomId, out MapLocality ml, out Dictionary<long, WorkstationStatus> ws) 
        {
            using (MiniZappaNewClient sc = new MiniZappaNewClient())
            {
                ml = sc.GetMaplocality(roomId);
                ws = sc.GetMapLocalityWorkstationStatuses(roomId);
            }
        }

        private void CreateDirectories(params string[] paths)
        {
            foreach (string path in paths)
            {
                if (!Directory.Exists(path))
                {
                    Directory.CreateDirectory(path);
                }
            }
        }

        #endregion

    }
}
